from abc import abstractmethod as _abstractmethod, ABCMeta as _ABCMeta
from socket import socket as _socket, AF_INET as _AF_INET, SOCK_STREAM as _SOCK_STREAM
from threading import Lock as _Lock, Thread as _Thread
from typing import Self as _Self


class Service(object, metaclass=_ABCMeta):
    def __init__(self, port: int) -> None:
        """
        :param port: the port on which the service listens
        """
        self._lock: _Lock = _Lock()
        self._port: int = port
        self._socket: _socket = _socket(_AF_INET, _SOCK_STREAM, proto=0)
        self._main_thread: _Thread | None = None

    def port(self) -> int:
        return self._port

    @_abstractmethod
    def run(self, *args, **kwargs) -> None:
        """
        Override this method to define the specific workflow.
        :param args: args
        :param kwargs: kwargs
        """
        raise NotImplementedError

    def _run(self, *args, **kwargs) -> None:
        """
        This method is equivalent to `Service.run()`. It leaves a middle layer for possible features in subclasses.
        :param args: args passed to `Service.run()`
        :param kwargs: kwargs passed to `Service.run()`
        """
        self.run(*args, **kwargs)

    def _register_process(self, *args, **kwargs) -> None:
        """
        Register the multithread worker.
        :param args: args passed to `Service.run()`
        :param kwargs: kwargs passed to `Service.run()`
        """
        self._lock.acquire()
        if self._main_thread:
            raise RuntimeWarning("A service can only run once")
        try:
            self._main_thread = _Thread(name=f"service{hash(self)}", target=self._run, args=args, kwargs=kwargs)
        finally:
            self._lock.release()

    def _parallel_run(self, *args, **kwargs) -> None:
        """
        This method is similar to `Service._run()` except that it runs the workflow in a child thread.
        :param args: args passed to `Service.run()`
        :param kwargs: kwargs passed to `Service.run()`
        """
        self._register_process(*args, **kwargs)
        self._main_thread.start()

    def start(self, parallel: bool = False, *args, **kwargs) -> _Self:
        """
        This is the publicly exposed interface to start the service.
        :param parallel: True: main thread not blocked; False: main thread blocked
        :param args: args passed to `Service.run()`
        :param kwargs: kwargs passed to `Service.run()`
        :return: self
        """
        try:
            return self
        finally:
            if parallel:
                self._parallel_run(*args, **kwargs)
            else:
                self._run(*args, **kwargs)

    @_abstractmethod
    def kill(self) -> None:
        raise NotImplementedError


class Connection(object):
    def __init__(self, service: Service, socket: _socket, address: tuple[str, int],
                 remainder_data: bytes = b"") -> None:
        """
        :param service: the service to which it belongs
        :param socket: the connection socket
        :param address: peer address
        :param remainder_data: message remain from last connection
        """
        self._service: Service = service
        self._socket: _socket = socket
        self._address: tuple[str, int] = address
        self._remainder: bytes = remainder_data

    def __str__(self) -> str:
        return self._address[0] + ":" + str(self._address[1])

    def closed(self) -> bool:
        """
        :return: True: the socket is closed; False: the socket is active
        """
        return self._socket.fileno() == -1

    def _require_open_socket(self, mandatory: bool = True) -> _socket:
        """
        Check if the socket is active and return it
        :param mandatory: True: an open socket is required; False: a closed socket is acceptable
        :return: the socket object
        """
        if mandatory and self.closed():
            raise IOError("An open socket is required")
        return self._socket

    def receive(self, chunk_size: int = 512) -> bytes | None:
        """
        :param chunk_size: chunk buffer size
        :return: message in bytes or None
        """
        if self._remainder != b"":
            if (i := self._remainder.find(b";")) != len(self._remainder) - 1:
                msg = self._remainder[:i + 1]
                self._remainder = self._remainder[i:]
            else:
                msg = self._remainder[:i]
                self._remainder = b""
            return msg
        try:
            msg = b""
            while not msg.endswith(b";"):
                msg += self._require_open_socket().recv(chunk_size)
            if (i := msg.find(b";")) != len(msg) - 1:
                self._remainder = msg[i + 1:]
                return msg[:i]
            return msg[:len(msg) - 1]
        except IOError:
            return

    def send(self, msg: bytes) -> None:
        """
        :param msg: message in bytes
        """
        self._require_open_socket().send(msg + b";")
        if msg == b"disconnect":
            self.close()

    def disconnect(self) -> None:
        """
        Request disconnection.
        """
        self.send(b"disconnect")

    def close(self) -> None:
        """
        Directly close the socket.
        """
        self._require_open_socket(False).close()


class Callback(object):
    def on_initialize(self, service: Service) -> None:
        pass

    def on_fail(self, service: Service, error: Exception) -> None:
        pass

    def on_connect(self, service: Service, connection: Connection) -> None:
        pass

    def on_receive(self, service: Service, msg: bytes) -> None:
        pass

    def on_disconnect(self, service: Service) -> None:
        pass


class Entity(Service, metaclass=_ABCMeta):
    def __init__(self, port: int, callback: Callback) -> None:
        """
        :param port: the port on which the service listens
        :param callback: the callback interface
        """
        super().__init__(port)
        self.callback: Callback = callback

    def _stage(self, connection: Connection) -> None:
        while True:
            msg = connection.receive()
            if not msg or msg == b"disconnect":
                self.callback.on_disconnect(self)
                return connection.close()
            self.callback.on_receive(self, msg)

    def _run(self, *args, **kwargs) -> None:
        try:
            return super()._run(*args, **kwargs)
        except Exception as e:
            self.callback.on_fail(self, e)



from leads.controller.controller import *
from leads.controller.device import *
from leads.controller.input import *
from leads.controller.output import *
from leads.controller.predefined_tags import *



from datetime import datetime
from time import time, sleep

from PySimpleGUI import Button, Text, Column
from keyboard import add_hotkey

from leads import *
from leads.comm import *
from leads_dashboard import *
from leads_vec.__version__ import __version__
from leads_vec.config import Config


class CustomRuntimeData(RuntimeData):
    m1_mode: int = 0
    m3_mode: int = 0


def main(main_controller: Controller, config: Config) -> int:
    context = Leads(srw_mode=config.srw_mode)

    def render(manager: ContextManager):
        def switch_m1_mode():
            manager.rd().m1_mode = (manager.rd().m1_mode + 1) % 2

        def switch_m3_mode():
            manager.rd().m3_mode = (manager.rd().m3_mode + 1) % 3

        manager["m1"] = Button(font=("Arial", config.font_size_small), key=switch_m1_mode,
                               size=(round(manager.window().width() / 21), 13))
        manager["m2"] = Button(font=("Arial", config.font_size_x_large),
                               size=(round(manager.window().width() / 126), 2))
        manager["m3"] = Button(font=("Arial", config.font_size_medium), key=switch_m3_mode,
                               size=(round(manager.window().width() / 42), 7))
        manager["dtcs_status"] = Text(text="DTCS READY", text_color="green", font=("Arial", config.font_size_small),
                                      size=(round(manager.window().width() / 40), None))
        manager["abs_status"] = Text(text="ABS READY", text_color="green", font=("Arial", config.font_size_small),
                                     size=(round(manager.window().width() / 40), None))
        manager["ebi_status"] = Text(text="EBI READY", text_color="green", font=("Arial", config.font_size_small),
                                     size=(round(manager.window().width() / 40), None))
        manager["atbs_status"] = Text(text="ATBS READY", text_color="green", font=("Arial", config.font_size_small),
                                      size=(round(manager.window().width() / 40), None))
        manager["comm_status"] = Text(text="COMM ONLINE", text_color="white", font=("Arial", config.font_size_small),
                                      size=(round(manager.window().width() / 40), None))

        def switch_dtcs():
            context.set_dtcs(not (dtcs_enabled := context.is_dtcs_enabled()))
            manager["dtcs"].update(f"DTCS {'OFF' if dtcs_enabled else 'ON'}")

        add_hotkey("1", switch_dtcs)

        def switch_abs():
            context.set_abs(not (abs_enabled := context.is_abs_enabled()))
            manager["abs"].update(f"ABS {'OFF' if abs_enabled else 'ON'}")

        add_hotkey("2", switch_abs)

        def switch_ebi():
            context.set_ebi(not (ebi_enabled := context.is_ebi_enabled()))
            manager["ebi"].update(f"EBI {'OFF' if ebi_enabled else 'ON'}")

        add_hotkey("3", switch_ebi)

        def switch_atbs():
            context.set_atbs(not (atbs_enabled := context.is_atbs_enabled()))
            manager["atbs"].update(f"ATBS {'OFF' if atbs_enabled else 'ON'}")

        add_hotkey("4", switch_atbs)

        manager["dtcs"] = Button(button_text="DTCS ON", key=switch_dtcs, font=("Arial", config.font_size_small),
                                 size=(round(manager.window().width() / 35), None))
        manager["abs"] = Button(button_text="ABS ON", key=switch_abs, font=("Arial", config.font_size_small),
                                size=(round(manager.window().width() / 35), None))
        manager["ebi"] = Button(button_text="EBI ON", key=switch_ebi, font=("Arial", config.font_size_small),
                                size=(round(manager.window().width() / 35), None))
        manager["atbs"] = Button(button_text="ATBS ON", key=switch_atbs, font=("Arial", config.font_size_small),
                                 size=(round(manager.window().width() / 35), None))

    uim = initialize(
        Window(config.width,
               config.height,
               config.refresh_rate,
               CustomRuntimeData(),
               fullscreen=config.fullscreen,
               no_title_bar=False),
        render,
        context,
        main_controller)

    class CustomCallback(Callback):
        def on_fail(self, service: Service, error: Exception) -> None:
            uim.rd().comm = None
            for _ in range(30):
                if uim.active():
                    break
                sleep(.1)
            uim["comm_status"].update("COMM OFFLINE", text_color="gray")

        def on_receive(self, service: Service, msg: bytes) -> None:
            print(msg)

    uim.rd().comm = start_client(config.comm_addr, create_client(config.comm_port, CustomCallback()), True)

    class CustomListener(EventListener):
        def on_push(self, e: DataPushedEvent) -> None:
            try:
                uim.rd().comm_notify(e.data)
            except IOError:
                uim.rd().comm_kill()
                uim.rd().comm = None
                uim["comm_status"].update("COMM OFFLINE", text_color="gray")

        def on_update(self, e: UpdateEvent) -> None:
            duration = int(time()) - uim.rd().start_time
            if uim.rd().m1_mode == 0:
                uim["m1"].update("LAP TIME\n\nLAP1 9s\nLAP2 11s\nLAP3 10s")
            elif uim.rd().m1_mode == 1:
                uim["m1"].update(f"VeC {__version__.upper()}\n\n"
                                 f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n"
                                 f"{duration // 60} MIN {duration % 60} SEC\n\n"
                                 f"{'SRW MODE' if config.srw_mode else 'DRW MODE'}\n"
                                 f"REFRESH RATE: {config.refresh_rate} FPS")
            uim["m2"].update(f"{int(context.data().front_wheel_speed)}")
            if uim.rd().m3_mode == 0:
                uim["m3"].update("0.0V")
            elif uim.rd().m3_mode == 1:
                uim["m3"].update("G Force")
            else:
                uim["m3"].update("Speed Trend")

        def on_intervene(self, e: InterventionEvent) -> None:
            uim[e.system.lower() + "_status"].update(e.system + " INTERVENED", text_color="purple")

        def post_intervene(self, e: InterventionEvent) -> None:
            uim[e.system.lower() + "_status"].update(e.system + " READY", text_color="green")

        def on_suspend(self, e: SuspensionEvent) -> None:
            uim[e.system.lower() + "_status"].update(e.system + " SUSPENDED", text_color="red")

        def post_suspend(self, e: SuspensionEvent) -> None:
            uim[e.system.lower() + "_status"].update(e.system + " READY", text_color="green")

    context.set_event_listener(CustomListener())
    uim.layout([
        [
            Column([[uim["m1"]]], element_justification='c', expand_x=True),
            Column([[uim["m2"]]], element_justification='c', expand_x=True),
            Column([[uim["m3"]]], element_justification='c', expand_x=True)
        ],
        [
            Column([[uim["dtcs_status"]]], element_justification='c', expand_x=True),
            Column([[uim["abs_status"]]], element_justification='c', expand_x=True),
            Column([[uim["ebi_status"]]], element_justification='c', expand_x=True),
            Column([[uim["atbs_status"]]], element_justification='c', expand_x=True),
            Column([[uim["comm_status"]]], element_justification='c', expand_x=True)
        ],
        [
            Column([[uim["dtcs"]]], element_justification='c', expand_x=True),
            Column([[uim["abs"]]], element_justification='c', expand_x=True),
            Column([[uim["ebi"]]], element_justification='c', expand_x=True),
            Column([[uim["atbs"]]], element_justification='c', expand_x=True)
        ]
    ])
    uim.show()
    uim.rd().comm_kill()
    return 0



from argparse import ArgumentParser as _ArgumentParser
from sys import exit as _exit

from leads_vec.config import load_config as _load_config, DEFAULT_CONFIG as _DEFAULT_CONFIG

if __name__ == '__main__':
    parser = _ArgumentParser(prog="LEADS",
                             description="Lightweight Embedded Assisted Driving System",
                             epilog="GitHub: https://github.com/ProjectNeura/LEADS")
    parser.add_argument("-c", "--config", nargs=1, default=None, help="configuration file")
    try:
        from leads_emulation import SRWRandom as _Controller
    except ImportError:
        raise ImportError("At least one adapter has to be installed")
    args = parser.parse_args()
    config = _load_config(args.config) if args.config else _DEFAULT_CONFIG
    from leads_vec.cli import main as _main

    _exit(_main(_Controller("main"), config))



from copy import copy as _copy
from typing import TextIO as _TextIO, TypeVar as _TypeVar, Generic as _Generic, Sequence as _Sequence, \
    Callable as _Callable

from numpy import mean as _mean

T = _TypeVar("T")

Compressor = _Callable[[list[T], int], list[T]]
Stringifier = _Callable[[T], str]


def mean_compressor(sequence: list[T], target_size: int) -> list[T]:
    """
    A compression method that reduces data memory usage by averaging adjacent numbers and merging them.
    :param sequence: the sequence to compress
    :param target_size: expected size
    :return: the compressed sequence
    """
    chunk_size = int(len(sequence) / target_size)
    if chunk_size < 2:
        return sequence
    r = []
    for i in range(target_size - 1):
        r.append(_mean(sequence[i * chunk_size: (i + 1) * chunk_size]))
    r.append(_mean(sequence[(target_size - 1) * chunk_size:]))
    return r


def csv_stringifier(element: T) -> str:
    """
    Dump an element as a CSV string.
    :param element: the element to stringify
    :return: CSV string
    """
    return str(element) + ","


class DataPersistence(_Sequence, _Generic[T]):
    def __init__(self,
                 file: str | _TextIO,
                 max_size: int = -1,
                 chunk_scale: int = 1,
                 compressor: Compressor = mean_compressor,
                 stringifier: Stringifier = csv_stringifier) -> None:
        """
        :param file: the file into which the data is written
        :param max_size: maximum cached size
        :param chunk_scale: chunk scaling factor (compression)
        :param compressor: compressor interface
        :param stringifier: stringifier interface
        """
        self._file: _TextIO = open(file, "a") if isinstance(file, str) else file
        self._max_size: int = max_size
        self._chunk_scale: int = chunk_scale
        self._compressor: Compressor = compressor
        self._stringifier: Stringifier = stringifier
        self._data: list[T] = []
        self._chunk: list[T] = []
        self._chunk_size: int = chunk_scale

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, item: int | slice) -> T | list[T]:
        return self._data[item]

    def __str__(self) -> str:
        return str(self._data)

    def close(self) -> None:
        self._file.close()

    def get_chunk_size(self) -> int:
        return self._chunk_size

    def to_list(self) -> list[T]:
        return _copy(self._data)

    def get_chunk(self) -> list[T]:
        return _copy(self._chunk)

    def _push_to_data(self, element: T) -> None:
        self._data.append(element)
        if self._max_size < 2:
            return
        if len(self._data) >= self._max_size:
            self._data = self._compressor(self._data, int(len(self._data) * .5))
            self._chunk_size *= 2

    def append(self, element: T) -> None:
        self._file.write(self._stringifier(element))
        if self._chunk_size == 1:
            return self._push_to_data(element)
        self._chunk.append(element)
        if len(self._chunk) >= self._chunk_size:
            for e in self._compressor(self._chunk, self._chunk_scale):
                self._push_to_data(e)
            self._chunk.clear()



from threading import Thread as _Thread
from time import sleep as _sleep
from typing import Callable as _Callable, Self as _Self, TypeVar as _TypeVar, Generic as _Generic

from PySimpleGUI import Window as _Window, Element as _Element, WINDOW_CLOSED as _WINDOW_CLOSED, theme as _theme

from leads_dashboard.runtime import RuntimeData


def default_on_refresh() -> None:
    pass


def default_on_kill() -> None:
    pass


Widget = _Element

T = _TypeVar("T", bound=RuntimeData)


class Window(_Generic[T]):
    def __init__(self,
                 width: int,
                 height: int,
                 refresh_rate: int,
                 runtime_data: T,
                 on_refresh: _Callable[[_Self], None] = default_on_refresh,
                 on_kill: _Callable[[_Self], None] = default_on_kill,
                 title: str = "LEADS",
                 fullscreen: bool = True,
                 no_title_bar: bool = True) -> None:
        _theme("Default1")
        self._root: _Window = _Window(title,
                                      size=(None if width < 0 else width, None if height < 0 else height),
                                      text_justification="center",
                                      no_titlebar=no_title_bar)
        self._width: int = _Window.get_screen_size()[0] if fullscreen else width
        self._height: int = _Window.get_screen_size()[1] if fullscreen else height
        self._fullscreen: bool = fullscreen
        self._refresh_rate: int = refresh_rate
        self._refresh_interval: float = float(1 / refresh_rate)
        self._runtime_data: T = runtime_data
        self._on_refresh: _Callable[[_Self], None] = on_refresh
        self._on_kill: _Callable[[_Self], None] = on_kill

        self._active: bool = False
        self._refresher_thread: _Thread | None = None

    def root(self) -> _Window:
        return self._root

    def width(self) -> int:
        return self._width

    def height(self) -> int:
        return self._height

    def refresh_interval(self) -> float:
        return self._refresh_interval

    def refresh_rate(self) -> int:
        return self._refresh_rate

    def runtime_data(self) -> T:
        return self._runtime_data

    def set_on_refresh(self, on_refresh: _Callable[[_Self], None]) -> None:
        self._on_refresh = on_refresh

    def set_on_close(self, on_close: _Callable[[_Self], None]) -> None:
        self._on_kill = on_close

    def active(self) -> bool:
        return self._active

    def refresher(self) -> None:
        while self._active:
            self._root.write_event_value("refresher", None)
            self._runtime_data.frame_counter += 1
            _sleep(self._refresh_interval)

    def show(self) -> None:
        self._root.finalize()
        if self._fullscreen:
            self._root.maximize()
        self._active = True
        self._refresher_thread = _Thread(name="refresher", target=self.refresher)
        self._refresher_thread.start()
        while self._active:
            event, values = self._root.read()
            if event == _WINDOW_CLOSED:
                self._active = False
                break
            elif event == "refresher":
                self._on_refresh(self)
            elif callable(event):
                event()

    def kill(self) -> None:
        self._active = False
        self._root.close()


class ContextManager(object):
    def __init__(self, window: Window) -> None:
        self._window: Window = window
        self._widgets: dict[str, Widget] = {}

    def __setitem__(self, key: str, widget: Widget) -> None:
        self._widgets[key] = widget

    def __getitem__(self, key: str) -> Widget:
        return self._widgets[key]

    def set(self, key: str, widget: Widget) -> None:
        self[key] = widget

    def get(self, key: str) -> Widget:
        return self[key]

    def parse_layout(self, layout: list[list[str | Widget]]) -> list[list[Widget]]:
        for i in range(len(layout)):
            for j in range(len(layout[i])):
                e = layout[i][j]
                if isinstance(e, str):
                    layout[i][j] = self[e]
        return layout

    def layout(self, layout: list[list[str | Widget]]) -> None:
        self._window.root().layout(self.parse_layout(layout))

    def window(self) -> Window:
        return self._window

    def rd(self) -> T:
        return self._window.runtime_data()

    def active(self) -> bool:
        return self._window.active()

    def root(self) -> _Window:
        return self._window.root()

    def show(self) -> None:
        self._window.show()

    def kill(self) -> None:
        self._window.kill()



from abc import abstractmethod as _abstractmethod, ABCMeta as _ABCMeta
from typing import TypeVar as _TypeVar, Generic as _Generic

from leads.controller.device import Device

T = _TypeVar("T")


class Input(Device, _Generic[T], metaclass=_ABCMeta):
    @_abstractmethod
    def read(self) -> T:
        raise NotImplementedError


class Sensor(Input[T], metaclass=_ABCMeta):
    pass



from importlib.util import find_spec as _find_spec

if not _find_spec("PySimpleGUI"):
    raise ImportError("Please install `pysimplegui` to run this module\n>>>pip install pysimplegui")

from typing import Callable as _Callable, TypeVar as _TypeVar

from leads import Leads as _Leads, Controller as _Controller
from leads.comm import Server as _Server, create_server as _create_server
from leads_dashboard.prototype import *
from leads_dashboard.runtime import *

T = _TypeVar("T")


def initialize(window: Window,
               render: _Callable[[ContextManager], None],
               leads: _Leads[T],
               main_controller: _Controller[T]) -> ContextManager:
    ctx = ContextManager(window)
    render(ctx)
    window.runtime_data().frame_counter = 0

    def on_refresh(_):
        leads.push(main_controller.collect_all())
        leads.update()

    window.set_on_refresh(on_refresh)
    return ctx


def initialize_comm_server(window: Window,
                           render: _Callable[[ContextManager], None],
                           server: _Server = _create_server()) -> None:
    ctx = ContextManager(window)
    render(ctx)
    server.start(True)

    def on_close(_):
        server.kill()

    window.set_on_close(on_close)



from abc import abstractmethod as _abstractmethod, ABCMeta as _ABCMeta
from typing import TypeVar as _TypeVar, Generic as _Generic

from leads.controller.device import Device

T = _TypeVar("T")


class Output(Device, _Generic[T], metaclass=_ABCMeta):
    @_abstractmethod
    def write(self, payload: T) -> None:
        raise NotImplementedError



from typing import TypeVar as _TypeVar, Any as _Any

from leads.constant import SYSTEM_DTCS, SYSTEM_ABS, SYSTEM_EBI, SYSTEM_ATBS
from leads.context import Context
from leads.event import EventListener, DataPushedEvent, UpdateEvent, SuspensionEvent, InterventionEvent, \
    InterventionExitEvent

T = _TypeVar("T")

_OptionalNumber = int | float | None


def dtcs_srw(context: Context,
             front_wheel_speed: _OptionalNumber,
             rear_wheel_speed: _OptionalNumber) -> InterventionEvent:
    if rear_wheel_speed and front_wheel_speed < rear_wheel_speed:
        return InterventionEvent(context, SYSTEM_DTCS, front_wheel_speed, rear_wheel_speed)
    return InterventionExitEvent(context, SYSTEM_DTCS, front_wheel_speed, rear_wheel_speed)


def dtcs_drw(context: Context,
             front_wheel_speed: _OptionalNumber,
             left_rear_wheel_speed: _OptionalNumber,
             right_rear_wheel_speed: _OptionalNumber) -> InterventionEvent:
    if left_rear_wheel_speed and front_wheel_speed < left_rear_wheel_speed:
        return InterventionEvent(context, SYSTEM_DTCS, "l", front_wheel_speed, left_rear_wheel_speed)
    if right_rear_wheel_speed and front_wheel_speed < right_rear_wheel_speed:
        return InterventionEvent(context, SYSTEM_DTCS, "r", front_wheel_speed, right_rear_wheel_speed)
    return InterventionExitEvent(context, SYSTEM_DTCS, front_wheel_speed, left_rear_wheel_speed, right_rear_wheel_speed)


class Leads(Context[T]):
    def __init__(self, event_listener: EventListener = EventListener(), *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._event_listener: EventListener = event_listener

    def set_event_listener(self, event_listener: EventListener) -> None:
        self._event_listener = event_listener

    def _acquire_data(self, name: str, *systems: str, mandatory: bool = True) -> _Any | None:
        try:
            return self.data().__getattribute__(name)
        except AttributeError:
            if mandatory:
                for system in systems:
                    self._event_listener.on_suspend(SuspensionEvent(self, system, f"no data for `{name}`"))

    def push(self, data: T) -> None:
        self._event_listener.on_push(DataPushedEvent(self, data))
        super().push(data)
        self._event_listener.post_push(DataPushedEvent(self, data))

    def intervene(self, event: InterventionEvent) -> None:
        if isinstance(event, InterventionExitEvent):
            self._event_listener.post_intervene(event)
        else:
            self._event_listener.on_intervene(event)

    def update(self) -> None:
        self._event_listener.on_update(UpdateEvent(self))

        if front_wheel_speed := self._acquire_data("front_wheel_speed",
                                                   SYSTEM_DTCS,
                                                   SYSTEM_ABS,
                                                   SYSTEM_EBI,
                                                   SYSTEM_ATBS):
            rear_wheel_speed = self._acquire_data("rear_wheel_speed",
                                                  SYSTEM_DTCS, SYSTEM_ATBS,
                                                  mandatory=self.in_srw_mode())
            left_rear_wheel_speed = self._acquire_data("left_rear_wheel_speed",
                                                       SYSTEM_DTCS, SYSTEM_ATBS,
                                                       mandatory=not self.in_srw_mode())
            right_rear_wheel_speed = self._acquire_data("right_rear_wheel_speed",
                                                        SYSTEM_DTCS, SYSTEM_ATBS,
                                                        mandatory=not self.in_srw_mode())
            # DTCS
            if self.is_dtcs_enabled():
                if self.in_srw_mode():
                    self.intervene(dtcs_srw(self, front_wheel_speed, rear_wheel_speed))
                else:
                    self.intervene(dtcs_drw(self, front_wheel_speed, left_rear_wheel_speed, right_rear_wheel_speed))

        self._event_listener.post_update(UpdateEvent(self))
